use crate::read;
use itertools::Itertools;

#[derive(Debug, PartialEq, Eq, Clone, Hash, Default)]
struct Coord {
    pub row: usize,
    pub col: usize,
}

pub fn run() {
    let input = read("12");
    let lines = input.lines();
    let mut start: Coord = Default::default();
    let mut target: Coord = Default::default();
    let map = lines
        .enumerate()
        .map(|(row, l)| {
            l.chars()
                .enumerate()
                .map(|(col, c)| match c {
                    'S' => {
                        start = Coord {
                            row,
                            col,
                        };
                        0
                    }
                    'E' => {
                        target = Coord {
                            row,
                            col,
                        };
                        25
                    }
                    c => (c as u8) - ('a' as u8),
                })
                .collect_vec()
        })
        .collect_vec();

    #[cfg(feature = "part1")]
    {
        println!("Day 12, Part 01: {:?}", bfs(&map, start, target.clone()).unwrap());
    }

    #[cfg(feature = "part2")]
    {
        let min = map
            .iter()
            .enumerate()
            .flat_map(|(row, v)| {
                v.iter()
                    .enumerate()
                    .filter(|(_, vv)| **vv == 0)
                    .map(|(col, _)| Coord {
                        row,
                        col,
                    })
                    .collect_vec()
            })
            .filter_map(|start| bfs(&map, start.clone(), target.clone()))
            .min()
            .unwrap();
        println!("Day 12, Part 02: {:?}", min);
    }
}

fn bfs(map: &Vec<Vec<u8>>, start: Coord, target: Coord) -> Option<usize> {
    let mut visited = vec![vec![false; map[0].len()]; map.len()];
    let mut queue = std::collections::VecDeque::new();
    queue.push_back((start, 0));

    while let Some((next, path_length)) = queue.pop_front() {
        if next == target {
            return Some(path_length);
        }
        for coord in [(1, 0), (-1, 0), (0, 1), (0, -1)]
            .iter()
            .filter_map(|(dr, dc)|make_coord((next.row as isize) + dr, (next.col as isize) + dc, &map)) {
            if map[next.row][next.col] + 1
                < map[coord.row][coord.col] || visited[coord.row][coord.col] {
                continue;
            }
            visited[coord.row][coord.col] = true;
            queue.push_back((coord, path_length + 1));
        }
    }
    None
}

fn make_coord(row: isize, col: isize, map: &Vec<Vec<u8>>) -> Option<Coord> {
    if row >= 0
        && col >= 0
        && row < map.len() as isize
        && (col as usize) < map[row as usize].len() {
        Some(Coord {
            row: row as usize,
            col: col as usize
        })
    } else {
        None
    }
}
